/*
 * Mp3Splt -- Utility for mp3/ogg splitting without decoding
 *
 * Copyright (c) 2002-2003 M. Trotta - <matteo.trotta@lib.unimib.it>
 *
 * http://mp3splt.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Tested on: Linux 2.4.18 #4 Fri May 31 01:25:31 PDT 2002 i686
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "mp3.h"
#include "splt.h"
#include "mad.h"

#if !HAVE_FSEEKO
#define fseeko fseek
#define ftello ftell
#endif

#define MAD_BSIZE 2016

// frequence table
const int freqs[9] = { 44100, 48000, 32000,
			 22050, 24000, 16000 ,
			 11025 , 12000 , 8000 };

// bitrate table
const int tabsel_123[2][3][16] = {
  { {128,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
    {128,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
    {128,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

  { {128,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
    {128,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
    {128,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};

const char categories[GENRENUM][10] = {
	{"blues"},
	{"classical"},
	{"country"},
	{"data"},
	{"folk"},
	{"jazz"},
	{"newage"},
	{"reggae"},
	{"rock"},
	{"soundtrack"},
	{"misc"},
};

const char unsigned id3genre[GENRENUM] = {0x00, 0x20, 0x02, 0xFF, 0x50, 0x08, 0x0a, 0x10, 0x11, 0x18, 0xFF};

// Fill out ID3 buffer
char *id3 (char *id, char *title, char *artist, char *album, char *year, unsigned char genre, char *comment, unsigned char track)
{
	char buffer[30];
	int j=3,i;
	strncpy(id, TAG, 3);

	memset(buffer, '\0', 30);
	if (title!=NULL) strncpy(buffer, title, strlen(title));
	for (i=0; i<30; i++) id[j++]=buffer[i];

	memset(buffer, '\0', 30);
	if (artist!=NULL) strncpy(buffer, artist, strlen(artist));
	for (i=0; i<30; i++) id[j++]=buffer[i];

	memset(buffer, '\0', 30);
	if (album!=NULL) strncpy(buffer, album, strlen(album));
	for (i=0; i<30; i++) id[j++]=buffer[i];

	memset(buffer, '\0', 30);
	if (year!=NULL) strncpy(buffer, year, strlen(year));
	for (i=0; i<4; i++) id[j++]=buffer[i];

	memset(buffer, '\0', 30);
	if (comment!=NULL) strncpy(buffer, comment, strlen(comment));
	for (i=0; i<30; i++) id[j++]=buffer[i];
	if (track!=0x00)
		id[j-1] = (char) track;
	id[j]=(char) genre;

	return id;
}

unsigned char getgenre (char *s)
{
	int i;
	for (i=0; i<GENRENUM; i++) {
		if (strncmp(s, categories[i], strlen(s))==0)
			return id3genre[i];
	}
	return 0xFF;
}

void checksync (mp3_state *state)
{
	char junk[32];
	fprintf(stderr, "\nWarning: Too many sync errors! This may not be a mp3 file. Continue? (y/n) ");
	fgets(junk, 31, stdin);
	if (junk[0]=='y')
		state->syncdetect = 0;
	else error("Aborted.");
}

// Calculates bitrate
int c_bitrate (unsigned long head)
{
    if ((head & 0xffe00000) != 0xffe00000) return 0;
    if (!((head>>17)&3)) return 0;
    if (((head>>12)&0xf) == 0xf) return 0;
    if (!((head >> 12) & 0xf)) return 0;
    if (((head>>10)&0x3) == 0x3 ) return 0;
    if (((head >> 19) & 1)==1 && ((head>>17)&3)==3 && ((head>>16)&1)==1) return 0;
    if ((head & 0xffff0000) == 0xfffe0000) return 0;

    return ((head>>12)&0xf);
}

int c_freq (int mpgid, int idex, unsigned long head)
{
	int srate;
	if(!idex)
	   srate = 6 + ((head>>10)&0x3);
	else srate = ((head>>10)&0x3) + ((1-mpgid)*3);
	return freqs[srate];
}

struct header makehead (unsigned long headword, struct mp3 mp3f, struct header head, off_t ptr)
{
	head.ptr = ptr;
	head.bitrate = tabsel_123[1 - mp3f.mpgid][mp3f.layer-1][c_bitrate(headword)];
	head.padding = ((headword>>9)&0x1);
	head.framesize = (head.bitrate*144000)/(mp3f.freq<<(1 - mp3f.mpgid)) + head.padding;

	return head;
}

// Finds first header from start_pos. Returns -1 if no header is found
off_t findhead (mp3_state *state, off_t start)
{
	if (getword(state->file_input, start, SEEK_SET, &state->headw) == -1)
		return -1;
	if (feof(state->file_input)) return -1;
	while (!(c_bitrate(state->headw))) {
		if (feof(state->file_input)) return -1;
		state->headw <<= 8;
		state->headw |= fgetc(state->file_input);
		start++;
	}
	return start;
}

// Finds first valid header from start. Will work with high probabilty, i hope :)
off_t findvalidhead (mp3_state *state, off_t start)
{
	off_t begin;
	struct header h;
	
	begin = findhead(state, start);
	do {
		start = begin;
		if (start == -1) break;
		h = makehead (state->headw, state->mp3file, h, start);
		begin = findhead(state, (start + 1));
	} while (begin!=(start + h.framesize));
	
	return start;
}

int checkvbr (unsigned long head1, unsigned long head2)
{
	if (c_bitrate(head1)==c_bitrate(head2))
		return 0;
	else return 1;
}

int getid3v1 (FILE *file_input)
{
	if (fseeko(file_input, (off_t) -128, SEEK_END)==-1)
		return 0;
	if (fgetc(file_input)=='T')
	   if (fgetc(file_input)=='A')
	      if (fgetc(file_input)=='G')
			return -128;
	return 0;
}

// Check if there is a ID3v2. If found, it returns offset of mp3 data.
off_t getid3v2 (FILE *in, off_t start)
{
	unsigned long oword = 0;
	if (fseeko(in, start, SEEK_SET)==-1)
		return 0;
	if (fgetc(in)=='I')
		if (fgetc(in)=='D')
			if (fgetc(in)=='3') {
				int i;
				if (fseeko(in, (off_t) 3, SEEK_CUR)==-1)
					return 0;
				for (i=0; i<4; i++)
					oword = (oword << 7) | fgetc(in);
				return (off_t) (oword);
			}

	return 0;
}

// Fill out mp3 informations
struct mp3 fillmp3info (mp3_state *state)
{
	if (state->mp3file.firsthead.ptr!=-1) {
		state->mp3file.mpgid = (state->headw >> 19)&1;
		state->mp3file.idex = (state->headw >> 20)&1;
		state->mp3file.freq = c_freq(state->mp3file.mpgid, state->mp3file.idex, state->headw);
		state->mp3file.layer = 4-((state->headw>>17)&3);
		state->mp3file.firsthead = makehead(state->headw, state->mp3file, state->mp3file.firsthead, state->mp3file.firsthead.ptr);
		state->mp3file.channels = ((state->headw >> 6)&3);
	}
	return state->mp3file;
}

int xing_header(mp3_state *state, off_t start, int len)
{
	unsigned long headw;
	int i;
	
	if (getword(state->file_input, start, SEEK_SET, &headw) == -1)
		return 0;

	for (i=4; i<len; i++)
	{
		if ((headw==0x58696E67) || (headw==0x496E666F)) // "Xing" or "Info"
			return i;
		headw <<= 8;
		headw |= fgetc(state->file_input);
	}

	return 0;
}

mp3_state *mp3info(FILE *file_input, mp3_state *state, int quietoption, int framemode)
{
	off_t begin;
	unsigned long headw;

	if ((state = (mp3_state *) malloc (sizeof(mp3_state)))==NULL) {
		perror("malloc");
		exit(1);
	}

	if (quietoption)
		state->syncdetect = 0;
	else state->syncdetect = 1;
	state->frames = 0;
	state->end = 0;
	state->syncerrors = 0;
	state->file_input = file_input;
	state->framemode = framemode;
	state->headw = 0;
	state->mp3file.xing = 0;
	state->mp3file.xingbuffer = NULL;
	state->mp3file.len = flength(file_input);
	
	state->silence_list = NULL;
	
	begin = findhead(state, getid3v2 (state->file_input, 0));
	do {
		state->mp3file.firsthead.ptr = begin;
		if (state->mp3file.firsthead.ptr==-1) {
			mp3_state_free(state);
			return NULL;
		}
		state->mp3file = fillmp3info (state);
		begin = findhead(state, state->mp3file.firsthead.ptr + 1);
	} while (begin!=(state->mp3file.firsthead.ptr + state->mp3file.firsthead.framesize));

	// Check if VBR file
	if ((state->mp3file.xing_offset = xing_header(state, state->mp3file.firsthead.ptr, state->mp3file.firsthead.framesize))>0) {
		state->mp3file.xing = state->mp3file.firsthead.framesize;
		if ((state->mp3file.xingbuffer = (unsigned char *) malloc(state->mp3file.xing))==NULL) {
			perror("malloc");
			exit(1);
		}
		memset(state->mp3file.xingbuffer, 0x00, state->mp3file.xing);
		if (fseeko(state->file_input, state->mp3file.firsthead.ptr, SEEK_SET)==-1) {
			mp3_state_free(state);
			return NULL;
		}
		fread(state->mp3file.xingbuffer, sizeof(char), state->mp3file.xing, state->file_input);
		if (getword(state->file_input, state->mp3file.firsthead.ptr+state->mp3file.xing_offset+4, SEEK_SET, &state->frames) == -1) {
			mp3_state_free(state);
			return NULL;
		}
		state->mp3file.firsthead.ptr += state->mp3file.firsthead.framesize;
		if (getword(state->file_input, state->mp3file.firsthead.ptr, SEEK_SET, &headw) == -1) {
			mp3_state_free(state);
			return NULL;
		}
		state->mp3file.firsthead = makehead (headw, state->mp3file, state->mp3file.firsthead, state->mp3file.firsthead.ptr);
	}

	if (!state->framemode) {
		if (state->mp3file.xing==0) {
			if (getword(state->file_input, state->mp3file.firsthead.ptr, SEEK_SET, &headw) == -1) {
				mp3_state_free(state);
				return NULL;
			}
			if (begin==-1)
				error("file is too short!");
		}

		if ((state->mp3file.xing!=0) || (checkvbr(state->headw, headw))) {
			if (!quietoption)
				fprintf (stderr, "Warning: I think mp3 is VBR or has Info header. Switching to Frame mode...\n");
			state->framemode = 1;
		}
	}
	
	return state;
}

int mp3split (unsigned char *filename, mp3_state *state, char *id3, float fbegin_sec, float fend_sec, short adjustoption, int threshold)
{
	off_t begin = 0, end = -1;
	short ret, eof=0;
	
	if (state->framemode) {

		unsigned long fbegin, fend, adjust=0;
		
		fbegin = fbegin_sec * state->mp3file.fps;
		if (fend_sec != -1)
			fend = fend_sec * state->mp3file.fps;
		else fend = 0xFFFFFFFF;

		if (state->end == 0) {
			state->h.ptr = state->mp3file.firsthead.ptr;
			state->h.framesize = state->mp3file.firsthead.framesize;
			begin = state->mp3file.firsthead.ptr;
			// Finds begin by counting frames
			while (state->frames < fbegin) {
				begin = findhead(state, state->h.ptr + state->h.framesize);
				if (begin==-1)
					return -1;
				if ((begin!=state->h.ptr + state->h.framesize)&&(state->syncerrors>=0)) state->syncerrors++;
				if ((state->syncdetect)&&(state->syncerrors>MAXSYNC))
					checksync(state);
				state->h = makehead (state->headw, state->mp3file, state->h, begin);
				state->frames++;
			}
		}
		else
			begin = state->end;
			
		if (state->mp3file.len > 0)
			if (begin >= state->mp3file.len) // If we can check, we just do that :)
				return -1;

		// Finds end by counting frames
		if (adjustoption) {
			if (fend_sec != -1) {
				unsigned long len;
				adjust = (unsigned long) (adjustoption * state->mp3file.fps);
				len = (fend-fbegin)-1;
				if (adjust > len)
					adjust = len;
				if (fend > adjust)
					fend -= adjust;
			}
			else adjust=0;
		}
		
		while (state->frames < fend) {
			state->frames++;
			end = findhead(state, state->h.ptr + state->h.framesize);
			if (end == -1) {
				end = state->h.ptr + state->h.framesize; // Last valid offset
				eof=1;
				break;
			}
			if ((end!=state->h.ptr + state->h.framesize)&&(state->syncerrors>=0)) state->syncerrors++;
			if ((state->syncdetect)&&(state->syncerrors>MAXSYNC))
				checksync(state);
			state->h = makehead (state->headw, state->mp3file, state->h, end);
			if ((adjust) && (state->frames >= fend)) {
				if (mp3_scan_silence(state, end, 2 * adjust, threshold, 0) > 0)
					adjust = (short) (state->silence_list->position * state->mp3file.fps);
				fend += adjust;
				end = findhead(state, end);
				ssplit_free(&state->silence_list);
				adjust=0;
			}
		}

		state->end = end;

		if (state->mp3file.xing!=0) {
			unsigned long headw;
			headw = (unsigned long) (state->frames - fbegin + 1); // Frames
			state->mp3file.xingbuffer[state->mp3file.xing_offset+4] = (headw >> 24) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+5] = (headw >> 16) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+6] = (headw >> 8) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+7] = headw  & 0xFF;
			if (end == -1)
				end = state->mp3file.len;
			headw = (unsigned long) (end - begin + state->mp3file.xing); // Bytes
			state->mp3file.xingbuffer[state->mp3file.xing_offset+8] = (headw >> 24) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+9] = (headw >> 16) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+10] = (headw >> 8) & 0xFF;
			state->mp3file.xingbuffer[state->mp3file.xing_offset+11] = headw  & 0xFF;
		}
	}
	else {
		// If file is CBR we can trust bitrate to find position
		if (state->end == 0) {
			begin = (off_t) (fbegin_sec * state->mp3file.bitrate + state->mp3file.firsthead.ptr);
			// Finds first valid header. Mantain clean files.
			begin = findvalidhead (state, begin);
			if (begin==-1)
				return -1;
		}
		else begin = state->end;
		if (fend_sec != -1) {
			end = (off_t) (fend_sec * state->mp3file.bitrate + state->mp3file.firsthead.ptr);
			end = findvalidhead (state, end); // We take the complete frame
		}
		state->end = end;
	}
	
	ret = split (filename, state->file_input, begin, end, state->mp3file.xing, state->mp3file.xingbuffer, id3);
	
	if (eof) return -3;
	else return ret;
}
 
off_t adjustsync(mp3_state *state, off_t begin, off_t end)
{
	off_t position;
	position = begin;
	if (fseeko(state->file_input, position, SEEK_SET)==-1)
		return (off_t) (-1);
	
	while (position++ < end) { // First we search for ID3v1
		if (fgetc(state->file_input)=='T') {
			if (fgetc(state->file_input)=='A') {
				if (fgetc(state->file_input)=='G')
					return (position + 127);
				else position++;
			}
			fseeko(state->file_input, -1, SEEK_CUR);
		}
	}
	
	position = begin;
	if (fseeko(state->file_input, position, SEEK_SET)==-1)
		return (off_t) (-1);
	
	while (position++ < end) { // Now we search for ID3v2
		if (fgetc(state->file_input)=='I') {
			if (fgetc(state->file_input)=='D') {
				if (fgetc(state->file_input)=='3')
					return (position - 1);
				else position++;
			}
			fseeko(state->file_input, -1, SEEK_CUR);
		}
	}
	
	return end;
}

int syncerror_split (mp3_state *state, int quiet, char *dir)
{
	off_t splitpoints[MAXTRACKS+1], offset = 0;
	char junk[512];
	int i, ret;
	
	state->h.ptr = state->mp3file.firsthead.ptr;
	state->h.framesize = state->mp3file.firsthead.framesize;
	splitpoints[0] = 0;
	
	fprintf(stderr, "Processing file to detect possible split points, please wait...\n");
	
	while (state->syncerrors < MAXTRACKS) {
		offset = findhead(state, state->h.ptr + state->h.framesize);
		if (offset==-1)
			break;
		if (offset!=state->h.ptr + state->h.framesize) {
			state->syncerrors++;
			splitpoints[state->syncerrors] = adjustsync(state, state->h.ptr, offset);
			if (splitpoints[state->syncerrors] == -1)
				return -1;
			offset = findvalidhead(state, splitpoints[state->syncerrors]);
			if (getword(state->file_input, offset, SEEK_SET, &state->headw) == -1)
				return -1;
		}
		state->h = makehead (state->headw, state->mp3file, state->h, offset);
	}
	
	if (state->syncerrors == 0)
		return 1;
	if (state->syncerrors == MAXTRACKS)
		return 2;

	splitpoints[++state->syncerrors] = -1;

	fprintf(stderr, "Total tracks found: %ld\n", state->syncerrors);
	if (!quiet) {
		fprintf(stderr, "Is this a reasonable number of tracks for this file? (y/n) ");
		fgets(junk, 31, stdin);
		if (junk[0]=='n')
			return 3;
	}

	memset(junk, 0x0, 512);
	
	for (i = 0; i < state->syncerrors; i++) {
		sprintf(junk, "Track %02d.mp3", i+1);
		if (dir)
		{
			char temp[512];
			strncpy(temp, junk, 512);
			sprintf(junk, "%s%c%s", dir, DIRCHAR, temp);
		}
		fprintf (stderr, "\n %3ld %% -> Splitting %s... ", ((i+1)*100)/state->syncerrors, junk);
		ret = split(junk, state->file_input, splitpoints[i], splitpoints[i+1], 0, NULL, NULL);
		if ((ret==0) || (ret==-3))
			fprintf (stderr, "OK");
		else {
			fprintf (stderr, "FAILED\n");
			return -1;
		}
	}
	
	fprintf(stderr, "\n\n");
	
	return 0;
}

int mp3_silence(struct mad_synth *synth, int channels, short threshold)
{
	int i, j;
	short sample;
	register int value;
	
	for (j=0; j<channels; j++) {
		for(i=0; i<synth->pcm.length; i++)
		{
			value = (int) (synth->pcm.samples[j][i] >> (MAD_F_FRACBITS-15));
			if(value > 32767)
				value=32767;
			if(value < -32768)
				value=-32768;
			if (value < 0)
				value = -value;

			sample = (short) (value);
			if (sample > threshold)
				return 0;
		}
	}
	return 1;
}

int mp3_scan_silence (mp3_state *state, off_t begin, unsigned long frames, short threshold, short output)
{
	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;
	int len = 0, found = 0, shot;
	short first, flush = 0;
	unsigned char inputBuffer[MAD_BSIZE];
	unsigned long frameCount=0, silence_begin = 0, silence_end = 0;
	off_t pos;
	
	pos = begin;
	
	if (fseeko(state->file_input, begin, SEEK_SET)==-1)
		return -1;
	
	if (frames == 0)
		frames = 0xFFFFFFFF;
		
	first = output;
	shot = DEFAULTSHOT;
		
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);
	
	if (output) fprintf(stderr, "[  0 %%]\r");
	
	do
	{
		
		if (feof(state->file_input))
			break;
		
		if(stream.buffer==NULL || stream.error==MAD_ERROR_BUFLEN)
		{
			size_t readSize, remaining;
			unsigned char *readStart;
		
			if(stream.next_frame!=NULL)
			{
				remaining = stream.bufend - stream.next_frame;
				memmove(inputBuffer, stream.next_frame, remaining);
				readStart = inputBuffer + remaining;
				readSize = MAD_BSIZE - remaining;
			}
			else {
				readSize=MAD_BSIZE;
				readStart=inputBuffer;
				remaining=0;
			}
			pos = ftello(state->file_input);
			readSize=fread(readStart, 1, readSize, state->file_input);
			if (readSize <= 0)
				break;
			mad_stream_buffer(&stream, inputBuffer, readSize+remaining);
			stream.error = MAD_ERROR_NONE;
		}

		if(mad_frame_decode(&frame,&stream)) {
			if(MAD_RECOVERABLE(stream.error))
				continue;
			else
				if(stream.error==MAD_ERROR_BUFLEN)
					continue;
				else
					break;
		}
		mad_synth_frame(&synth,&frame);
		frameCount++;
		if (frameCount >= frames)
			flush = 1;
			
		if ((!flush) && (mp3_silence(&synth, MAD_NCHANNELS(&frame.header), threshold))) {
			if (len == 0) silence_begin = frameCount-1;
			if (first == 0) len++;
			if (shot < DEFAULTSHOT)
				shot+=2;
			silence_end = frameCount-1;
		}
		else {
			if (len > DEFAULTSILLEN) {
				if ((flush) || (shot <= 0))
				 {
				 	float position;
					silence_begin = (silence_end - silence_begin);
					silence_begin >>= DEFAULTSILOFF;
					silence_end -= silence_begin;
					position = (float) (silence_end / state->mp3file.fps);
					ssplit_new(&state->silence_list, position, len);
					found++;
					len = 0;
					shot = DEFAULTSHOT;
				}
			}
			else len = 0;
			if ((first) && (shot <= 0))
				first = 0;
			if (shot > 0) shot--;
		}
		
		if ((output) && (state->mp3file.len > 0)) 
			if (frameCount%1000 == 0) fprintf(stderr, "[%3d %%]\r", (int)(pos/(state->mp3file.len/100)));
	
	} while ((frameCount < frames) && (found < MAXTRACKS));
	
	if (output) fprintf(stderr, "[100 %%]\n");
	
	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);
	
	return found;
}

void mp3_state_free (mp3_state *state)
{
	if (state->mp3file.xingbuffer)
		free(state->mp3file.xingbuffer);
	ssplit_free(&state->silence_list);
	free(state);
}
